//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#ifndef FACTOR_GNSS_TDCP_BATCH_H_
#define FACTOR_GNSS_TDCP_BATCH_H_

//Wolf includes
#include "core/common/wolf.h"
#include "core/factor/factor_autodiff.h"
#include "gnss/sensor/sensor_gnss.h"

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorGnssTdcpBatch);

class FactorGnssTdcpBatch : public FactorAutodiff<FactorGnssTdcpBatch, 4, 3, 4, 1, 3, 4, 1, 3, 1, 1, 1>
{
    protected:
        SensorGnssPtr sensor_gnss_ptr_;

    public:

        FactorGnssTdcpBatch(const FeatureBasePtr& _ftr_ptr,
                            const CaptureBasePtr& _capture_other_ptr,
                            const SensorGnssPtr& _sensor_gnss_ptr,
                            const ProcessorBasePtr& _processor_ptr,
                            bool _apply_loss_function = false,
                            FactorStatus _status = FAC_ACTIVE) :
            FactorAutodiff<FactorGnssTdcpBatch, 4, 3, 4, 1, 3, 4, 1, 3, 1, 1, 1>("FactorGnssTdcpBatch",
                                                                                 TOP_GEOM,
                                                                                 _ftr_ptr,
                                                                                 _capture_other_ptr->getFrame(),
                                                                                 _capture_other_ptr,
                                                                                 nullptr,
                                                                                 nullptr,
                                                                                 _processor_ptr,
                                                                                 _apply_loss_function,
                                                                                 _status,
                                                                                 _capture_other_ptr->getFrame()->getP(),
                                                                                 _capture_other_ptr->getFrame()->getO(),
                                                                                 _capture_other_ptr->getStateBlock('T'),
                                                                                 _ftr_ptr->getFrame()->getP(),
                                                                                 _ftr_ptr->getFrame()->getO(),
                                                                                 _ftr_ptr->getCapture()->getStateBlock('T'),
                                                                                 _sensor_gnss_ptr->getP(),
                                                                                 _sensor_gnss_ptr->getEnuMapRoll(),
                                                                                 _sensor_gnss_ptr->getEnuMapPitch(),
                                                                                 _sensor_gnss_ptr->getEnuMapYaw()),
            sensor_gnss_ptr_(_sensor_gnss_ptr)
        {
            assert(_ftr_ptr->getMeasurement().size() == 4 && "FactorGnssTdcpBatch uses 4d measurements (pos.displacement, delta clock). For FeatureGnssTdcp with only displacement, use FactorGnssTdcp3d instead");
            WOLF_WARN_COND(!sensor_gnss_ptr_->isEnuDefined(), "Creating a GNSS SingleDiff 3D factor without initializing ENU");
        }

        ~FactorGnssTdcpBatch() override = default;

        template<typename T>
        bool operator ()(const T* const _x1,
                         const T* const _o1,
                         const T* const _t1,
                         const T* const _x2,
                         const T* const _o2,
                         const T* const _t2,
                         const T* const _x_antena,
                         const T* const _roll_ENU_MAP,
                         const T* const _pitch_ENU_MAP,
                         const T* const _yaw_ENU_MAP,
                         T* _residuals) const;

};

template<typename T>
inline bool FactorGnssTdcpBatch::operator ()(const T* const _x1,
                                             const T* const _o1,
                                             const T* const _t1,
                                             const T* const _x2,
                                             const T* const _o2,
                                             const T* const _t2,
                                             const T* const _x_antena,
                                             const T* const _roll_ENU_MAP,
                                             const T* const _pitch_ENU_MAP,
                                             const T* const _yaw_ENU_MAP,
                                             T* _residuals) const
{
    Eigen::Map<const Eigen::Matrix<T,3,1>> t_MAP_BASE1(_x1);
    Eigen::Map<const Eigen::Quaternion<T>> q_MAP_BASE1(_o1);
    Eigen::Map<const Eigen::Matrix<T,3,1>> t_MAP_BASE2(_x2);
    Eigen::Map<const Eigen::Quaternion<T>> q_MAP_BASE2(_o2);
    Eigen::Map<const Eigen::Matrix<T,3,1>> t_BASE_ANTENA(_x_antena);
    Eigen::Map<Eigen::Matrix<T,3,1> > disp_residuals(_residuals);
    Eigen::Map<Eigen::Matrix<T,4,1> > residuals(_residuals);

    Eigen::Matrix<T,3,3> R_ECEF_ENU = sensor_gnss_ptr_->getREnuEcef().transpose().cast<T>();
    Eigen::Matrix<T,3,3> R_ENU_MAP = sensor_gnss_ptr_->computeREnuMap(_roll_ENU_MAP[0], _pitch_ENU_MAP[0], _yaw_ENU_MAP[0]);

    // Expected d
    Eigen::Matrix<T,4,1> exp;

    // Expected displacement in ECEF
    exp.head(3) = R_ECEF_ENU * R_ENU_MAP * ((q_MAP_BASE2 * t_BASE_ANTENA + t_MAP_BASE2) - (q_MAP_BASE1 * t_BASE_ANTENA + t_MAP_BASE1));

    // clock error
    exp(3) = *_t2 - *_t1;

    // Compute residual
    residuals = getMeasurementSquareRootInformationUpper().cast<T>() * (exp - getMeasurement().cast<T>());

    //std::cout << "frame1: " << _x1[0] << " " << _x1[1] << " " << _x1[2] << " " << _o1[0] << " " << _o1[1] << " " << _o1[2] << " " << _o1[3] << std::endl;
    //std::cout << "frame2: " << _x2[0] << " " << _x2[1] << " " << _x2[2] << " " << _o2[0] << " " << _o2[1] << " " << _o2[2] << " " << _o2[3] << std::endl;
    //std::cout << "antena: " << _x_antena[0] << " " << _x_antena[1] << " " << _x_antena[2] << std::endl;
    //std::cout << "RPY: " << _roll[0] << " " << _pitch[1] << " " << _yaw[2] << std::endl;
    //std::cout << "expected_ECEF: " << expected_ECEF[0] << " " << expected_ECEF[1] << " " << expected_ECEF[2] << std::endl;
    //std::cout << "getMeasurement(): " << getMeasurement().transpose << std::endl;

    return true;
}

} // namespace wolf

#endif
