//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include "gnss/factor/factor_gnss_single_diff_2d.h"
#include "gnss/feature/feature_gnss_single_diff.h"
#include "gnss/processor/processor_gnss_single_diff.h"
#include "gnss/capture/capture_gnss_single_diff.h"

namespace wolf
{

ProcessorGnssSingleDiff::ProcessorGnssSingleDiff(ParamsProcessorGnssSingleDiffPtr _params_gnss) :
        ProcessorBase("ProcessorGnssSingleDiff", 0, _params_gnss),
        params_gnss_(_params_gnss)
{
    //
}

ProcessorGnssSingleDiff::~ProcessorGnssSingleDiff()
{
    //
}

void ProcessorGnssSingleDiff::processCapture(CaptureBasePtr _capture)
{
    // TODO: keep captures in a buffer and deal with KFpacks

    //WOLF_DEBUG("ProcessorGnssSingleDiff::process()");
    incoming_capture_ = std::static_pointer_cast<CaptureGnssSingleDiff>(_capture);

    // discard capture with null or non-key origin frame
    if (incoming_capture_->getOriginFrame() == nullptr || !incoming_capture_->getOriginFrame()->isKey())
    {
        WOLF_WARN("process single difference with null frame origin, skipping...");
        return;
    }

    if (last_KF_ == nullptr)
        last_KF_ = incoming_capture_->getOriginFrame();

    // NEW KF? ------------------------------------------------
    FrameBasePtr new_frame = nullptr;

    // ALREADY CREATED KF
    PackKeyFramePtr KF_pack = buffer_pack_kf_.selectPack( incoming_capture_, params_->time_tolerance);
    if (KF_pack && KF_pack->key_frame != incoming_capture_->getOriginFrame())
    {
        new_frame = KF_pack->key_frame;
        WOLF_DEBUG( "PR ",getName()," - capture ", incoming_capture_->id(), " appended to existing KF " , KF_pack->key_frame->id() , " TS: ", KF_pack->key_frame->getTimeStamp());
    }
    // MAKE KF
    else if (voteForKeyFrame() && permittedKeyFrame())
    {
        new_frame = getProblem()->emplaceKeyFrame( incoming_capture_->getTimeStamp());
        getProblem()->keyFrameCallback(new_frame, shared_from_this(), params_->time_tolerance);
        WOLF_DEBUG( "PR ",getName()," - capture ", incoming_capture_->id(), " appended to new KF " , new_frame->id() , " TS: ", new_frame->getTimeStamp());
    }

    // ESTABLISH FACTOR ------------------------------------------------
    if (new_frame)
    {
        // LINK CAPTURE
        _capture->link(new_frame); // Add incoming Capture to the new Frame

        // EMPLACE FEATURE
        auto ftr = FeatureBase::emplace<FeatureGnssSingleDiff>(incoming_capture_, incoming_capture_->getData(),incoming_capture_->getDataCovariance());

        // ADD FACTOR
        emplaceFactor(ftr);

        // store last KF
        last_KF_ = new_frame;
    }

    // INITIALIZE ENU_MAP IF 4 NECESSARY CONDITIONS ------------------------------------------------
    //      1 - ENU-ECEF defined
    //      2 - not initialized
    //      3 - current capture in key-frame with factor
    //      4 - frames constained by the factor separated enough ( > enu_map_init_dist_min)
    if ( sensor_gnss_->isEnuDefined() &&
        !sensor_gnss_->isEnuMapInitialized() &&
         new_frame != nullptr &&
         incoming_capture_->getFrame() != nullptr && incoming_capture_->getFrame()->isKey() &&
         incoming_capture_->getData().norm() > params_gnss_->enu_map_init_dist_min)
    {
        WOLF_DEBUG("initializing enu map");
        sensor_gnss_->initializeEnuMapYaw(incoming_capture_->getOriginFrame()->getState(),
                                          incoming_capture_->getFrame()->getState(),
                                          incoming_capture_->getData());
    }
}

FactorBasePtr ProcessorGnssSingleDiff::emplaceFactor(FeatureBasePtr _ftr)
{
    //WOLF_DEBUG("creating the factor...");
    // 2d
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssSingleDiff2d>(_ftr, _ftr, incoming_capture_->getOriginFrame(), sensor_gnss_, shared_from_this(), params_->apply_loss_function);
    // 3d TODO
    else
        std::runtime_error("Single Differences in 3d not implemented yet.");

    return nullptr;
}

bool ProcessorGnssSingleDiff::voteForKeyFrame() const
{
    if (last_KF_==nullptr)
        return true;

    std::cout << "params_gnss_->time_th = " << params_gnss_->time_th << std::endl;
    std::cout << "(last_KF_->getTimeStamp() - incoming_capture_->getTimeStamp()) = " << (last_KF_->getTimeStamp() - incoming_capture_->getTimeStamp()) << std::endl;

    // Depending on time since the last KF with gnssfix capture
    if ((incoming_capture_->getTimeStamp() - last_KF_->getTimeStamp()) > params_gnss_->time_th)
        return true;

    // Distance criterion
    std::cout << "params_gnss_->dist_traveled = " << params_gnss_->dist_traveled << std::endl;
    Eigen::Vector2d v_current_origin = (sensor_gnss_->getREnuMap().transpose() * sensor_gnss_->getREnuEcef() * incoming_capture_->getData()).head<2>();
    std::cout << "v_current_origin: " << v_current_origin.transpose() << std::endl;
    Eigen::Vector2d v_origin_last_KF = last_KF_->getP()->getState() - incoming_capture_->getOriginFrame()->getP()->getState();
    std::cout << "v_origin_last_KF: " << v_origin_last_KF.transpose() << std::endl;
    std::cout << "v_current_origin + v_origin_last_KF: " << (v_current_origin + v_origin_last_KF).transpose() << std::endl;
    if ((v_current_origin + v_origin_last_KF).norm() > params_gnss_->dist_traveled)
        return true;

    // TODO: more alternatives?

    // otherwise
    return false;
}

bool ProcessorGnssSingleDiff::storeKeyFrame(FrameBasePtr _frame_ptr)
{
    return true;
}
bool ProcessorGnssSingleDiff::storeCapture(CaptureBasePtr _cap_ptr)
{
    return false;
}
void ProcessorGnssSingleDiff::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

ProcessorBasePtr ProcessorGnssSingleDiff::create(const std::string& _unique_name, const ParamsProcessorBasePtr _params)
{
    ProcessorGnssSingleDiffPtr prc = std::make_shared<ProcessorGnssSingleDiff>(std::static_pointer_cast<ParamsProcessorGnssSingleDiff>(_params));
    prc->setName(_unique_name);
    return prc;
}

} // namespace wolf


// Register in the FactorySensor
#include "core/processor/factory_processor.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR(ProcessorGnssSingleDiff)
} // namespace wolf
